"""Utilities Module. With several functions that are repeatedly used"""
import os
import socket
import spidev

try:
    import RPi.GPIO as GPIO
except RuntimeError:
    print("Error importing RPi.GPIO! Use sudo / run sudo usermod -aG gpio <myusername> to get permission")


def discover_i2c():
    """
    Scans address space and ports to discover connected I2C or SPI devices
    uses os i2cdetect for the 1 I2C port and also scans two SPI ports

    """
    print("Scan for connected I2C devices' addresses:")
    print(os.popen("i2cdetect -y 1").read())
    print("Checking if SPI Module is loaded:")
    print(os.popen("lsmod | grep spi").read())


def read_analogue(channel, spi_device=0, baud=1350000):
    """
    Reads an analogue signal from the connected SPI ADC device and returns channel reading.

    :param channel: ADC channel where sensor is connected.
    :type channel: int
    :param spi_device: Either 0 or 1 as there are only 2 spi ports
    :type spi_device: int
    :param baud: the bit rate, measured in bit/s clock rate used for device
    :type baud: int
    :return: Raw 1024 bit ADC output data.
    """
    spi = spidev.SpiDev()
    spi.open(0, spi_device)
    spi.max_speed_hz = baud  # spi clock speed
    _, byte1, byte2 = spi.xfer2([1, (8+channel) << 4, 0])
    raw_data = ((byte1 & 3) << 8) + byte2
    return raw_data


def switch_actuator(gpio_pin, state):
    """
    Function to switch actuator ON or OFF

    :param gpio_pin: The pin the fan relay (motor in demo) is connected to.
    :type gpio_pin: int
    :param state: Boolean indicating whether fan is on or off.
    :type state: boolean
    """
    GPIO.setup(gpio_pin, GPIO.OUT)  # repetitive, will need to be done once.
    GPIO.output(gpio_pin, state)


def scan_network():
    """
    Ger the IP address other than the loopback IP that the device has been allocated by DHCP
    Scan subnet /24 of IP address to check for LAN brokers' availability.

    :return: list of online devices responding to ICMP echo request using ping.
    """

    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect(('10.255.255.255', 1))
        ip = s.getsockname()[0]
    except:
        ip = '127.0.0.1'  # localhost loopback IP if not connected to wifi
    finally:
        s.close()  # close socket

    net = ip.split(".")[:-1]
    online_dev = list()
    print(f"Scanning Hosts {ip}'s Network:")
    for host in range(1, 0x64):  # 0xff only scanning /24 subnet
        dev = ".".join(net) + "." + str(host)
        response = os.popen(f"ping -w 1 {dev}")  # -c 1
        try:
            if response.readlines()[5]:
                print(f"{dev} is online")
                online_dev.append(dev)
        except IndexError:
            continue
    return online_dev


def find_broker():
    """
    Scan for online MQTT brokers then scan within network by checking online hosts then scanning for
    open MQTT ports

    :return: No return
    """

    online_dev = scan_network()
    online_dev = ["mqtt.eclipse.org", "test.mosquitto.org"]+online_dev  # prepend online test brokers
    for ip in online_dev:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # unencrypted port 1883 or encrypted port 8883 open.
        conn = min(s.connect_ex((ip, 1883)), s.connect_ex((ip, 8883)))
        if conn == 0:
            print(f'MQTT Port 1883 or 8883 OPEN for {ip}')
        else:
            print(f'MQTT Port 1883 or 8883 CLOSED for {ip}')
        s.close()
    print("Scan Completed")


def gpio_init():
    """
    Function to initialize the GPIO pins, numbering system used and communication protocols.
    GPIO.BCM IS THE DEFAULT

    """
    GPIO.setmode(GPIO.BCM)  # Physical Board Pin Numbers
    GPIO.setwarnings(False)


def cleanup():
    """
    GPIO.cleanup() and exit(0) for a graceful exit.

    """
    spidev.SpiDev().close()  # close SPI connection
    GPIO.cleanup()
    exit(0)

def create_sensor_type(read, write, config):  # Add support for new sensor type
    pass

def sensor_attach_i2c(SensorType1, addr, sample_rate, broker):  # Add sensor, assign broker and topic
    pass



"""Top-level package for SmartAgro."""
#this file is always run each time the package is imported.
#will use this to  effect automatic importing of modules. instead of smartagro.smartagro.func()

import smartagro.utils
import smartagro.smart
#from smartagro.smart import SmartAgro

#implement as: from smartagro.smartagro import * OR import smatagro then use smartagro.smartagro.func()

# from smartagro import *, smart.foo(), utils.bar(), smart.SmartAgro() .

__author__ = """Kudzai Chris Kateera"""
__email__ = 'kckateera@gmail.com'
__version__ = '0.2.0'
__all__ = ['smart', 'utils'] #imports for from <package_name> import *

print(f"Invoking __init.py__ for {__name__}")



"""Main module."""
from smartagro import utils
import paho.mqtt.client as mqtt
import adafruit_dht
import time


class SmartAgro(object):
    """
    Implemented After searching for a broker
    Instantiates an object which has sensors added to it then configures a broker.
    Sensors are attached added with corresponding topics
    Sensor Data is published and Actuator can be activated
    """
    def __init__(self):
        """
        Object constructor for the sensors and actuators to be attached to it.

        """
        self.client = None
        self.sensors = set()
        self.actuators = set()
        self.dhtDevice = adafruit_dht.DHT11(18)  # DHT init GPIO 18 (Physical 12)
        # if none, scan network for brokers and connect to identified broker.
        # scan with utils.find_broker()
        self.config_broker()
        utils.gpio_init()

    def config_broker(self, broker="test.mosquitto.org", qos=0, port=1883, stream_schema="json"):
        """
        Function to configure a new broker to be published to.

        :param broker: The url or ip address of the broker.
        :param qos: quality of service determining how many times message is sent. 0,1,2
        :param port: broker port in use. default 1883, ssl 8883
        :param stream_schema: the data stream schema used. Default is json
        :return: mqtt client object

        """
        self.client = mqtt.Client("RPi0-ZA-2020")  # create new client
        self.client.connect(broker, port)  # connect to broker
        self.client.publish(topic="smartagro/test", payload="Test Successful", qos=qos)  # TOPIC & test payload
        self.client.subscribe("smartagro/actuator/#")
        self.client.on_connect = self.on_connect
        self.client.on_message = self.on_message
        self.client.loop_start()

    @staticmethod
    def on_connect(client, userdata, flags, rc):
        """
        The callback for when a connection is established with the server.

        :param client: Mqtt Client
        :param userdata: Authentication data
        :param flags: Connection indicators
        :param rc: status code of connection
        """
        print("Connected with result code " + str(rc))

    @staticmethod
    def on_message(client, userdata, msg):
        """
        The callback for when a PUBLISH message is received from the server.

        :param client: MQtt client
        :param userdata: data used for authenticated connections
        :param msg: received message topic and payload in bytes
        """
        if msg.topic == "smartagro/actuator/GPIO15":
            utils.switch_actuator(int(msg.topic[-2:]), bool(int(msg.payload.decode('utf-8'))))

    def read_sensor(self, channel):
        """
        Reads sensor, publishes topic to broker, adds to active sensors

        :param channel: ADC channel to be read.
        """
        reading = utils.read_analogue(channel)/7
        reading = round(reading, 2)
        self.client.publish(f"smartagro/sensor/{'Moisture' if channel==0 else 'Light'}", reading)  # sensor ID
        self.sensors.add(f"smartagro/sensor/{'Moisture' if channel==0 else 'Light'}")
        return reading

    def get_dht(self):
        """
        A function to get readings from the single wire DHT11 device.

        :return: Temperature and Humidity Readings
        """
        try:
            temp = humidity = 0
            temp, humidity = self.dhtDevice.temperature, self.dhtDevice.humidity
        except RuntimeError as error:
            print(error.args[0])
            time.sleep(2.0)  # Retry after 2 seconds
            temp, humidity = self.dhtDevice.temperature, self.dhtDevice.humidity
        except Exception as error:
            self.dhtDevice.exit()
            raise error
        finally:
            yield round(temp, 2), round(humidity, 2)  # use generators

    def read_dht(self):
        temperature, humidity = next(self.get_dht())
        self.client.publish(f"smartagro/sensor/humidity", humidity)
        self.client.publish(f"smartagro/sensor/temperature", temperature)
        self.sensors.add(f"smartagro/sensor/humidity")
        self.sensors.add(f"smartagro/sensor/temperature")
        return temperature, humidity

    def read_all(self):
        """
        A function to read all the values at once

        :return: A list of current moisture, light, temperature, humidity values
        """
        moist = self.read_sensor(0)  # Moisture
        light = self.read_sensor(2)  # Light
        temp, humid = self.read_dht()  # DHT
        return moist, light, temp, humid

    def activate_actuator(self, gpio_pin, state):
        """
        A function to activate or deactivate an actuator.

        :param gpio_pin: GPIO pic of connected actuator.
        :type gpio_pin: int
        :param state: State whether it is on or Off
        :type state: bool
        """
        utils.switch_actuator(gpio_pin, state)
        self.client.publish(f"smartagro/actuator/GPIO{gpio_pin}", state)  # actuator ID
        self.actuators.add(f"smartagro/actuator/GPIO{gpio_pin}")

    def remove_device(self, device):
        """
        Function to remove device from published topics

        :param device: Device Topic
        :type device: str
        """
        try:
            self.actuators.remove(device) if device in self.actuators else self.sensors.remove(device)
        except KeyError:
            print("The device you are trying to remove des not exist. Check connected devices again!")

    def active_devices(self):
        return self.actuators | self.sensors



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages
import subprocess,sys

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['Click>=7.0', 'paho-mqtt==1.5.1', 'spidev', 'sockets', 'RPi.GPIO', 'adafruit-circuitpython-dht']
# production program requirements. eg ADC libraries, etc that i will use.
# requirements has debug n dev tools etc what u will include in venv

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3', ]

setup(
    author="Kudzai Chris Kateera",
    author_email='kckateera@gmail.com',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="Smart, Iot-enabled greenhouse monitoring and control API",
    entry_points={
        'console_scripts': [
            'smartagro=smartagro.cli:main',
        ],
    },
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='smartagro',
    name='smartagro',
    packages=find_packages(include=['smartagro', 'smartagro.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/chris-kck/smartagro',
    version='0.2.0',
    zip_safe=False,
)
